use crate::{fuse_fs::FuseFs, PathExt, MOUNT_OPTIONS};

use btfproto::server::FsProvider;
use btlib::{
    BlockPath, Result,
};
use fuse_backend_rs::{
    api::server::Server,
    transport::{self, FuseSession, FuseChannel},
};
use log::error;
use std::path::PathBuf;
use std::{
    collections::HashMap,
    ffi::{c_char, c_int, CString},
    fs::File,
    os::{fd::FromRawFd, unix::ffi::OsStrExt},
    path::Path,
    sync::Arc,
};
use tokio::{
    task::JoinSet,
};
use futures::future::FutureExt;

pub use private::FuseDaemon;

mod private {
    use super::*;

    #[link(name = "fuse3")]
    extern "C" {
        /// Opens a channel to the kernel.
        fn fuse_open_channel(mountpoint: *const c_char, options: *const c_char) -> c_int;
    }

    /// Calls into libfuse3 to mount this file system at the given path. The file descriptor to use
    /// to communicate with the kernel is returned.
    fn mount_at<P: AsRef<Path>>(mnt_point: P) -> File {
        let mountpoint = CString::new(mnt_point.as_ref().as_os_str().as_bytes()).unwrap();
        let options = CString::new(MOUNT_OPTIONS).unwrap();
        let raw_fd = unsafe { fuse_open_channel(mountpoint.as_ptr(), options.as_ptr()) };
        unsafe { File::from_raw_fd(raw_fd) }
    }

    pub struct FuseDaemon {
        set: JoinSet<()>,
    }

    impl FuseDaemon {
        const FSNAME: &str = "btfuse";
        const FSTYPE: &str = "bt";
        const NUM_THREADS: usize = 8;

        pub fn new<T: AsRef<Path>, P: 'static + FsProvider>(
            path: T,
            uid_map: HashMap<u32, Arc<BlockPath>>,
            fallback_path: Arc<BlockPath>,
            provider: P,
        ) -> Result<Self> {
            path.try_create_dir()?;
            let server = Arc::new(Server::new(FuseFs::new(provider, uid_map, fallback_path)));
            let session = Self::session(path)?;
            let mut set = JoinSet::new();
            for _ in 0..Self::NUM_THREADS {
                let server = server.clone();
                let channel = session.new_channel()?;
                let future = tokio::task::spawn_blocking(move || Self::server_loop(server, channel));
                let future = future.map(|result| {
                    if let Err(err) = result {
                        error!("server_loop produced an error: {err}");
                    }
                });
                set.spawn(future);
            }
            Ok(Self { set })
        }

        pub fn mnt_path<T: AsRef<Path>>(path: T) -> PathBuf {
            path.as_ref().join("mnt")
        }

        fn session<T: AsRef<Path>>(path: T) -> Result<FuseSession> {
            let mnt_path = Self::mnt_path(path);
            mnt_path.try_create_dir()?;
            let mut session =
                FuseSession::new(mnt_path.as_ref(), Self::FSNAME, Self::FSTYPE, false)?;
            session.set_fuse_file(mount_at(mnt_path));
            Ok(session)
        }

        /// Opens a channel to the kernel and processes messages received in an infinite loop.
        ///
        /// # Warning
        /// Because the future returned by this method is passed to `unsafe_cast`, you must not
        /// hold any type which is not `Send` and `Sync` across an `await` point.
        fn server_loop<P: 'static + FsProvider>(
            server: Arc<Server<FuseFs<P>>>,
            mut channel: FuseChannel,
        ) {
            loop {
                match channel.get_request() {
                    Ok(Some((reader, writer))) => {
                        // Safety: reader and writer are not mutated while the future returned from
                        // async_handle_message is alive.
                        if let Err(err) = server.handle_message(reader, writer.into(), None, None) {
                            error!("error while handling FUSE message: {err}");
                        }
                    }
                    Ok(None) => break,
                    Err(err) => {
                        match err {
                            // Occurs when the file system is unmounted.
                            transport::Error::SessionFailure(_) => break,
                            _ => error!("unexpected  error from Channel::get_request: {err}"),
                        }
                    }
                }
            }
        }

        pub async fn finished(&mut self) {
            while self.set.join_next().await.is_some() {}
        }
    }
}
